""" Support Muffin-Peewee. """
import peewee as pw
from muffin_peewee.models import to_simple
from wtforms import fields as f

from muffin_rest import RESTHandler, Form, RESTNotFound, Filter, default_converter


try:
    from wtfpeewee.orm import model_form, ModelConverter

    ModelConverter.defaults[pw.DateField] = f.DateField
    ModelConverter.defaults[pw.DateTimeField] = f.DateTimeField

    from muffin_peewee.fields import JSONField

    ModelConverter.defaults[JSONField] = f.TextAreaField

except ImportError:
    model_form = None


def pw_converter(handler, flt):
    """ Convert column name to filter. """
    return default_converter(handler, flt, PWFilter)


class PWRESTHandlerMeta(type(RESTHandler)):

    """ Peewee specific. """

    def __new__(mcs, name, bases, params):
        """ Prepare handler params. """
        model = params.get('model')
        params.setdefault('name', model and model._meta.db_table.lower() or name.lower())
        params.setdefault('model_pk', model and model._meta.primary_key or None)
        cls = super(PWRESTHandlerMeta, mcs).__new__(mcs, name, bases, params)
        if not cls.form and cls.model and model_form:
            cls.form = model_form(cls.model, base_class=Form, **cls.form_meta)
        return cls


class PWRESTHandler(RESTHandler, metaclass=PWRESTHandlerMeta):

    """ Support REST for Peewee. """

    filters_converter = pw_converter

    model = None
    model_pk = None

    # only, exclude, recurse
    simple_meta = {}

    # only, exclude, field_args
    form_meta = {}

    def get_many(self, request):
        """ Get collection. """
        return self.model.select()

    def get_one(self, request):
        """ Load a resource. """
        resource = request.match_info.get(self.name)
        if not resource:
            return None

        try:
            return self.collection.where(self.model_pk == resource).get()
        except Exception:
            raise RESTNotFound(reason='Resource not found.')

    def populate(self):
        """ Create object. """
        return self.model()

    def save_form(self, form, request, resource=None):
        """ Save data. """
        resource = yield from super(PWRESTHandler, self).save_form(
            form, request, resource=resource)
        resource.save()
        return resource

    def to_simple(self, data, many=False):
        """ Serialize data. """
        if many:
            return super(PWRESTHandler, self).to_simple(data, many)
        return to_simple(data, **self.simple_meta)

    def delete(self, request, resource=None):
        """ Delete a resource. """
        if resource is None:
            raise RESTNotFound(reason='Resource not found')
        resource.delete_instance()


class PWFilter(Filter):

    """ Base filter for Peewee handlers. """

    operations = dict(Filter.operations, **{
        '%': lambda a, b: a % b,
        '<<': lambda a, b: a << b,
    })

    def apply(self, query, value):
        """ Filter a query. """
        form_field = query.model_class._meta.fields.get(self.column_name)
        if not form_field:
            return query
        return query.where(self.op(form_field, value))


class PWMultiField(f.StringField):

    """ Support many values. """

    def process_formdata(self, valuelist):
        """ Save multivalues. """
        self.data = valuelist


class PWMultiFilter(PWFilter):

    """ Support multi values. """

    form_field = PWMultiField

    def apply(self, query, value):
        """ Filter a query. """
        form_field = query.model_class._meta.fields.get(self.column_name)
        if not form_field:
            return query

        return query.where(form_field << value)


class PWLikeFilter(PWFilter):

    """ Filter query by value. """

    def apply(self, query, value):
        """ Filter a query. """
        form_field = query.model_class._meta.fields.get(self.column_name)
        value = "*%s*" % value
        return query.where(form_field % value)



""" REST support. """
import datetime as dt

import muffin
from aiohttp import MultiDict
from muffin.handler import Handler, abcoroutine

from muffin_rest import RESTNotFound, RESTBadRequest, FILTER_PREFIX, default_converter, FilterForm


class RESTHandler(Handler):

    """ Implement handler for REST operations. """

    form = None
    filters = ()
    filters_converter = default_converter

    def __init__(self):
        """ Initialize filters. """
        self.filters_form = FilterForm(prefix=FILTER_PREFIX)
        for flt in self.filters:
            self.filters_converter(flt).bind(self.filters_form)

    @classmethod
    def connect(cls, app, *paths, methods=None, name=None, **kwargs):
        """ Connect to the application. """
        if not paths:
            paths = [muffin.sre('/%s(/{%s})?/?' % (cls.name, cls.name))]

        if name is None:
            name = "rest-%s" % cls.name

        return super(RESTHandler, cls).connect(app, *paths, methods=methods, name=name, **kwargs)

    @abcoroutine
    def dispatch(self, request, view=None):
        """ Process REST. """
        self.auth = yield from self.authorize(request)
        self.collection = yield from self.get_many(request)

        if request.method == 'POST':
            return (yield from super(RESTHandler, self).dispatch(request, view=view))

        resource = yield from self.get_one(request)

        if request.method == 'GET' and resource is None:

            # Filter collection
            self.collection = yield from self.filter(request)

        return (
            yield from super(RESTHandler, self).dispatch(request, resource=resource, view=view))

    @abcoroutine
    def authorize(self, request):
        """ Base point for authorization. """
        return True

    @abcoroutine
    def parse(self, request):
        """ Ensure that request.data is multidict. """
        data = yield from super().parse(request)
        if not isinstance(data, MultiDict):
            data = MultiDict(data)
        return data

    @abcoroutine
    def get_many(self, request):
        """ Base point for collect data. """
        return []

    @abcoroutine
    def filter(self, request):
        """ Filter collection. """
        return self.filters_form.process(self.collection, request.GET)

    @abcoroutine
    def get_one(self, request):
        """ Base point load resource. """
        return request.match_info.get(self.name)

    @abcoroutine
    def get_form(self, request, resource=None):
        """ Base point load resource. """
        formdata = yield from self.parse(request)

        if not self.form:
            raise muffin.MuffinException("%s.form is not defined." % type(self).__name__)

        if resource:
            data = {}
            for name, field, *args in (self.form._unbound_fields or self.form()._fields):
                value = getattr(resource, name, None)
                if isinstance(value, (dt.datetime, dt.date)):
                    field = field.bind(None, name, _meta=1)
                    value = value.strftime(field.format)
                data[name] = value

            data.update(formdata)
            formdata = MultiDict(data)

        return self.form(formdata, obj=resource)

    @abcoroutine
    def save_form(self, form, request, resource=None):
        """ Save self form. """
        if resource is None:
            resource = self.populate()

        form.populate_obj(resource)
        return resource

    def populate(self):
        """ Create object. """
        return object()

    def to_simple(self, data, many=False):
        """ Convert resource to simple. """
        if many:
            return [self.to_simple(r) for r in data]
        return str(data)

    @abcoroutine
    def get(self, request, resource=None):
        """ Get collection of resources. """
        if resource:
            return self.to_simple(resource)

        return self.to_simple(self.collection, True)

    @abcoroutine
    def post(self, request):
        """ Create a resource. """
        form = yield from self.get_form(request)
        if not form.validate():
            raise RESTBadRequest(json=form.errors)
        resource = yield from self.save_form(form, request)
        return self.to_simple(resource)

    @abcoroutine
    def put(self, request, resource=None):
        """ Update a resource. """
        if resource is None:
            raise RESTNotFound(reason='Resource not found')

        form = yield from self.get_form(request, resource=resource)
        if not form.validate():
            raise RESTBadRequest(json=form.errors)
        resource = yield from self.save_form(form, request, resource=resource)
        return self.to_simple(resource)

    patch = put

    @abcoroutine
    def delete(self, request, resource=None):
        """ Delete a resource. """
        if resource is None:
            raise RESTNotFound(reason='Resource not found')

    @classmethod
    def scheme(cls):
        """ Return self schema. """
        return {
            'form': cls.form and {
                name: str(type(field)) for (name, field) in cls.form()._fields.items()} or None,
            'methods': cls.methods,
            'description': cls.__doc__,
        }



""" Support admin filters. """

import operator
import wtforms as wtf


FILTER_PREFIX = 'mr-'


class FilterDefault:

    """ Default filters value. """

    def __str__(self):
        """ Nothing here. """
        return ""


class FilterForm(wtf.Form):

    """ Store filters for resource. """

    def process(self, collection, formdata=None, obj=None, data=None, **kwargs):
        """ Filter collection. """
        super(FilterForm, self).process(formdata, obj, data, **kwargs)
        self.filters = {}
        for field in self._fields.values():
            collection, active = field.flt.filter(collection, self.data)
            if active:
                self.filters[field.flt.column_name] = field.data
        return collection


class Filter:

    """ Implement filters. """

    form_field = wtf.StringField
    default = FilterDefault()
    options = {}
    operations = {
        '==': operator.eq,
        '!=': operator.ne,
        '>=': operator.ge,
        '<=': operator.le,
        '>': operator.gt,
        '<': operator.lt,
    }

    def __init__(self, column_name, filter_name=None, op='==', form_field=None, **options):
        """ Initialize a filter. """
        self.column_name = column_name
        self.filter_name = filter_name or column_name
        if form_field is not None and issubclass(form_field, wtf.Field):
            self.form_field = form_field
        self.options = options or self.options
        self.op = self.operations.get(op)

    def bind(self, form):
        """ Bind to filter's form. """
        form_field = self.form_field(**self.options)
        form_field = form._fields[self.filter_name] = form_field.bind(
            form, self.filter_name, prefix=form._prefix)
        form_field.flt = self

    def filter(self, collection, data):
        """ Load value and filter collection. """
        value = self.value(data)
        if value is self.default:
            return collection, False
        if self.op is None:
            return collection, True
        return self.apply(collection, value), True

    def value(self, data):
        """ Get value from data. """
        value = data.get(self.filter_name, self.default)
        return value or self.default

    def apply(self, collection, value):
        """ Filter collection. """
        return [o for o in collection if self.op(getattr(o, self.column_name, None), value)]


# Filters for base primitives
BoolFilter = type('BoolFilter', (Filter,), {'form_field': wtf.BooleanField})
IntegerFilter = type('IntegerFilter', (Filter,), {'form_field': wtf.IntegerField})
DateFilter = type('DateFilter', (Filter,), {'form_field': wtf.DateField})
DateTimeFilter = type('DateTimeFilter', (Filter,), {'form_field': wtf.DateTimeField})
ChoiceFilter = type('ChoiceFilter', (Filter,), {'form_field': wtf.SelectField})


def default_converter(handler, flt, fcls=Filter):
    """ Convert column name to filter. """
    if isinstance(flt, Filter):
        return flt

    if isinstance(flt, str):
        return fcls(flt)

    name, params = flt

    return fcls(name, **params)
